# -*- coding: utf-8 -*-
from __future__ import print_function, division
from pymic.net.cls.torch_pretrained_net import *

TorchClsNetDict = {
    'resnet18':   ResNet18,
    'vgg16':      VGG16,
    'mobilenetv2':MobileNetV2
}
